import octofuss
import os, os.path
import usersplugin.common as common
import octofuss.ldaputils as oldap
import octofuss.hooks
import ldap
import ldap.dn
import ldap.filter
import urllib.request, urllib.parse, urllib.error
import time

_ = lambda x:x

TIMEOUT = 100
FIRSTUID = 10000
FIRSTGID = 500
LASTUID = 64000
DEFAULT_GROUP = "Domain Users"

class UserDB(common.UserDB):
    # Scope is one of:
    def _search(self, base, scope, filter='(objectClass=*)', attrlist=None, attrsonly=0):
        self._connect()
        try:
            result_data = self.ldap.search_s(base, scope, filter, attrlist, attrsonly)
        except:
            self._connect()
            result_data = self.ldap.search_s(base, scope, filter, attrlist, attrsonly)
        for d in result_data:
            yield oldap.Object(*d)

    def _connect(self):
        self.ldap = ldap.initialize(self.conf_server)
        self.ldap.simple_bind_s(self.conf_binddn, self.conf_bindpw)

    def __init__(self, tree, conf):
        super(UserDB, self).__init__()
        self.tree = tree
        self.conf_server = None
        self.conf_binddn = None
        self.conf_bindpw = None
        self.conf_basedn = None
        if conf.has_section("LDAP"):
            if conf.has_option("LDAP", "host"):
                self.conf_server = conf.get("LDAP", "host")
                t = urllib.parse.splittype(self.conf_server)
                if t[0] == None:
                    conf_server = "ldap://"+self.conf_server

            if conf.has_option("LDAP", "binddn"):
                self.conf_binddn = conf.get("LDAP", "binddn")
            if conf.has_option("LDAP", "bindpw"):
                self.conf_bindpw = conf.get("LDAP", "bindpw")
            if conf.has_option("LDAP", "basedn"):
                self.conf_basedn = conf.get("LDAP", "basedn")

        if not self.conf_server: self.conf_server = 'ldap://localhost'
        if not self.conf_binddn: self.conf_binddn = 'cn=admin,dc=octofuss'
        if not self.conf_bindpw: self.conf_bindpw = ''
        if not self.conf_basedn: self.conf_basedn = "dc=octofuss"
        self.ldap = ldap.initialize(self.conf_server)
        self.ldap.simple_bind_s(self.conf_binddn, self.conf_bindpw)

        self.defaultGidNumber = None

        for g in self._search(base="cn=%s,ou=Groups,%s" % (ldap.dn.escape_dn_chars(DEFAULT_GROUP), self.conf_basedn), scope=ldap.SCOPE_BASE):
            self.defaultGidNumber = int(g["gidNumber"][0])
            #self.defaultGroupSID = g["sambaSID"][0]
            break
        if self.defaultGidNumber == None:
            raise Error("Default group '%s' not found" % DEFAULT_GROUP)
        self.localSambaSid = None



    def commitModify(self, obj):
        changes = [(x[1][0], x[0], list(map(str, x[1][1]))) for x in obj._changes.items()]
#        print "CHANGES", repr(changes)
        self._connect()
        self.ldap.modify_s(obj._dn, changes)
        obj._clear_changes()

    def list(self, group = False):
        """
        List all users.

        If a group is specified, list all users in that group
        """
        if group:
            # Instantiate the group
            group = self._want_group(group)
#            print group.data.memberUid
            # Iterate all the users that have gidNumber same as the one of the group
            for u in self._search(base="ou=Users,%s" % self.conf_basedn, scope=ldap.SCOPE_ONELEVEL, filter='(gidNumber=%d)' % int(group.data.gidNumber[0])):
                yield User(u)

            # Iterate all memberUid of the group
            for uid in group.data.memberUid:
                base = "uid=%s,ou=Users,%s" % (ldap.dn.escape_dn_chars(uid), self.conf_basedn)
                res = self._search(base="uid=%s,ou=Users,%s" % (ldap.dn.escape_dn_chars(uid), self.conf_basedn), scope=ldap.SCOPE_BASE)

                for u in res:
                    yield User(u)
#                    print "Done with", u._d
        else:
            base = "ou=Users,%s" % self.conf_basedn
            for u in self._search(base=base, scope=ldap.SCOPE_ONELEVEL):
                yield User(u)

    def listGroups(self):
        for g in self._search(base="ou=Groups,%s" % self.conf_basedn, scope=ldap.SCOPE_ONELEVEL):
            yield Group(g)

    def groups_of_user(self, user):
        """
        Generate the Group objects to which the given User object belongs to
        """
        user = self._want_user(user)
        # Primary group first
        pg = self.group_by_gid(user.lget(["gidNumber"]))
        if pg != None: yield pg
        for group in self._search(base="ou=Groups,%s" % self.conf_basedn, scope=ldap.SCOPE_ONELEVEL,
                                  filter="(memberUid=%s)" % user.uid):
            yield Group(group)

    def user(self, name):
        name = ldap.dn.escape_dn_chars(name)
        try:
            users = [x for x in self._search(base="uid=%s,ou=Users,%s" % (ldap.dn.escape_dn_chars(name), self.conf_basedn), scope=ldap.SCOPE_BASE)]
        except ldap.NO_SUCH_OBJECT:
            return None

        if len(users) == 0:
            return None
        else:
            return User(users[0])

    def group(self, name):
        name = ldap.dn.escape_dn_chars(name)
        try:
            groups = [x for x in self._search(base="cn=%s,ou=Groups,%s" % (name, self.conf_basedn), scope=ldap.SCOPE_BASE)]
        except ldap.NO_SUCH_OBJECT:
            return None

        if len(groups) == 0:
            return None
        else:
            return Group(groups[0])

    def group_by_gid(self, gid):
        gid = int(gid)
        groups = [x for x in self._search(base="ou=Groups,%s" % self.conf_basedn, scope=ldap.SCOPE_ONELEVEL,
                     filter="(gidNumber=%d)" % gid)]
        if len(groups) == 0:
            return None
        else:
            return Group(groups[0])

    def user_in_group(self, username, groupname):
        u = self.user(username)
        if u == None: return None
        g = self.group(groupname)
        if g == None: return None
        if g.data.gidNumber == u.data.gidNumber: return u
        if u.uid in g.data.memberUid: return u
        return None

    def group_in_user(self, username, groupname):
        u = self.user(username)
        if u == None: return None
        g = self.group(groupname)
        if g == None: return None
        if g.data.gidNumber == u.data.gidNumber: return g
        if u.uid in g.data.memberUid: return g
        return None

    def add_user_to_group(self, user, group):
        user = self._want_user(user)
        group = self._want_group(group)
        group.data.memberUid = group.data.memberUid + [user.uid]
        common.userDB.commitModify(group.data)
        return user, group

    def delete_user_from_group(self, user, group):
        user = self._want_user(user)
        group = self._want_group(group)
        self._connect()
        if user.uid in group.data.memberUid:
            self.ldap.modify_s(group.data._dn,
                [(ldap.MOD_DELETE, "memberUid", [str(user.uid)])])
        if user.data.gidNumber == group.data.gidNumber:
            user.data.gidNumber = [self.defaultGidNumber]
            user.data.sambaPrimaryGroupSID = [self._getUserSID(self.defaultGidNumber)]
            user.commit()

    def delete(self, username):
        u = self.user(username)
        if u == None:
            raise KeyError("User %s does not exist" % username)
        homeDir = u.data.homeDirectory[0]
        octofuss.hooks.invoke("delete_user", username, homeDir)
        self._connect()
        self.ldap.delete_s("uid=%s,ou=Users,%s" % (ldap.dn.escape_dn_chars(username), self.conf_basedn))
        # Delete users from memberUid of the various groups
        name = ldap.filter.escape_filter_chars(username)
        mod = (ldap.MOD_DELETE, "memberUid", [str(username)])
        for group in self._search(base="ou=Groups,%s" % self.conf_basedn, scope=ldap.SCOPE_ONELEVEL,
                     filter="(memberUid=%s)" % name):
            self._connect()
            self.ldap.modify_s(group._dn, [mod])
        if self.tree.has("/netperms/byuser/" + username):
            self.tree.delete("/netperms/byuser/"+username)
        common.UserLog(username=username, action="deleted")

    def delete_group(self, groupname):
        g = self.group(groupname)
        if g == None:
            raise KeyError("Group %s does not exist" % groupname)
        gid = str(g.get("gidNumber"))
        self._connect()
        self.ldap.delete_s("cn=%s,ou=Groups,%s" % (ldap.dn.escape_dn_chars(groupname), self.conf_basedn))
        mod = [(ldap.MOD_REPLACE, "gidNumber", [str(self.defaultGidNumber)]),
               (ldap.MOD_REPLACE, "sambaPrimaryGroupSID", [self._getUserSID(int(self.defaultGidNumber))])]
        for user in self._search(base="ou=Users,%s" % self.conf_basedn, scope=ldap.SCOPE_ONELEVEL,
                     filter="(gidNumber=%s)" % gid):
            self._connect()
            self.ldap.modify_s(user._dn, mod)

    def _get_free_uidnumber(self):
        # Retrieve all existing uids
        uids = []
        for u in self._search(base="ou=Users,%s" % self.conf_basedn, scope=ldap.SCOPE_ONELEVEL, attrlist = ["uidNumber"]):
            uid = int(u["uidNumber"][0])
            if uid < FIRSTUID: continue
            if uid > LASTUID: continue
            uids.append(uid)
        # Sort them so that we can find gaps
        uids.sort()
        if len(uids) == 0:
            return FIRSTUID
        # Look for the first gap
        last = uids[0]
        for u in uids[1:]:
            if u > last + 1:
                return last + 1
            last = u
        # Else, return the maximum one plus 1
        return uids[-1] + 1

    def _get_free_gidnumber(self):
        # Retrieve all existing gids
        gids = []
        for g in self._search(base="ou=Groups,%s" % self.conf_basedn, scope=ldap.SCOPE_ONELEVEL, attrlist = ["gidNumber"]):
            gid = int(g["gidNumber"][0])
            if gid < FIRSTGID: continue
            gids.append(gid)
        # Sort them so that we can find gaps
        gids.sort()
        if len(gids) == 0:
            return FIRSTGID
        # Look for the first gap
        last = gids[0]
        for u in gids[1:]:
            if u != last + 1:
                return last + 1
            last = u
        # Else, return the maximum one plus 1
        return gids[-1] + 1

    def _getUserSID(self, uidNumber):
        sidPrefix = self._searchLocalSid()
        sidPostfix = (int(uidNumber)*2) + 1000
        return sidPrefix+"-"+str(sidPostfix)

    def _searchLocalSid(self):
        """ Returns the samba sid of the server"""
        if self.localSambaSid:
            return self.localSambaSid
        for res in self._search(base="%s" % self.conf_basedn, scope=ldap.SCOPE_ONELEVEL, filter="(sambaDomainName=*)", attrlist=["sambaSID"]):
            self.localSambaSid = res["sambaSID"][0]
            break
        return self.localSambaSid

    def new_user(self, value):
        self._connect()
        data = self._fillInNewUser(value)
        user = value["uid"]
        self._connect()
        self.ldap.add_s(
            "uid=%s,ou=Users,%s" % (ldap.dn.escape_dn_chars(user), self.conf_basedn),
            list(data.items()),
        )
        octofuss.hooks.invoke("new_user", user,
            int(data["uidNumber"][0]),
            int(data["gidNumber"][0]),
            data["homeDirectory"][0])

        res = self.user(user)
        if "controllerGroup" in value:
            res.set("controllerGroup", value["controllerGroup"])
        common.UserLog(username=user, action="created")
        return res

    def new_group(self, value):
        self._connect()
        data = self._fillInNewGroup(value)
        group = value["cn"]
        self.ldap.add_s(
            "cn=%s,ou=Groups,%s" % (ldap.dn.escape_dn_chars(group), self.conf_basedn),
            [x for x in list(data.items()) if len(x[1]) > 0],
        )
        return self.group(group)

def _getfirstor(list, default):
    if len(list) > 0: return list[0]
    return default

class User(common.User):
    def commit(self):
        common.userDB.commitModify(self.data)

    def getControllerGroup(self):
        try:
            # Extract the information from the file system
            homedir = self.data.homeDirectory[0]
            info = os.stat(homedir)
            if (info.st_mode & 0o20) == 0:
                # If it is not group writable, then there is no controller group
                return None
            else:
                # Else, get the group name from the gid
                group = common.userDB.group_by_gid(info.st_gid)
                if group == None:
                    return str(info.st_gid)
                else:
                    return group.name
        except OSError:
            return None

    def setControllerGroup(self, value):
        # Call hooks with the numerical group ID so that they will still work if the
        # local Unix user database is not hooked into LDAP
        if value == "":
            octofuss.hooks.invoke("set_controller_group", self.uid, "", self.data.homeDirectory[0], common.userDB.defaultGidNumber)
        else:
            group = common.userDB.group(value)
            if group == None:
                raise KeyError("%s is not a valid group" % value)
            octofuss.hooks.invoke("set_controller_group", self.uid, group.data.gidNumber[0], self.data.homeDirectory[0], common.userDB.defaultGidNumber)

    controllerGroup = property(getControllerGroup, setControllerGroup, doc="FUSS controller group")

class Group(common.Group):
    def commit(self):
        common.userDB.commitModify(self.data)
